"""
KeeMeta provides the metaclass for the 'worktoy.keenum' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ._kee_desc import Base, MRO, Members, IsRoot
from ..core import Object
from ..mcls import BaseMeta
from ..waitaminute import TypeException, attributeErrorFactory
from ..waitaminute.keenum import KeeNameError, KeeIndexError
from ..waitaminute.keenum import KeeMemberError, KeeValueError
from . import KeeSpace as KSpace, Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias, Self, Iterator

  from . import KeeNum

  Bases: TypeAlias = tuple[type, ...]


class KeeMeta(BaseMeta):
  """
  KeeMeta provides the metaclass for the 'worktoy.keenum' module.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __allow_instantiation__ = False
  __num_members__ = None
  __is_root__ = None

  #  Public Variables
  isRoot = IsRoot()
  base = Base()
  mroNum = MRO()
  members = Members()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kw: Any) -> KSpace:
    """Prepares the namespace for the class."""
    bases = (*[b for b in bases if b.__name__ != '_InitSub'],)
    return KSpace(mcls, name, bases, **kw)

  def __new__(mcls, name: str, bases: Bases, space: KSpace, **kw) -> Self:
    """
    Creates a new instance of the class.
    This method is called when the class is created.
    """
    cls = super().__new__(mcls, name, bases, space, **kw)
    cls.__num_members__ = []
    if not [b for b in bases if b is not Object]:
      return cls
    cls.__allow_instantiation__ = True

    for i, (key, kee) in enumerate(space.__enumeration_members__.items()):
      self = None
      for base in bases:
        self = getattr(base, key, None)
      if self is None:
        self = cls(kee, )
      setattr(cls, key, self)
      cls.__num_members__.append(self)
    cls.__allow_instantiation__ = False
    return cls

  def __call__(cls, *args, **kwargs) -> Any:
    """
    Prevents instantiation of the class.
    This method is called when the class is instantiated.
    """
    if cls.__allow_instantiation__:
      return super().__call__(*args, **kwargs)
    return cls._resolveMember(args[0])

  def __getitem__(cls, identifier: Any) -> Any:
    """
    Gets a member of the enumeration by index or key.
    This method is used to get a member of the enumeration by index or key.
    """
    return cls._resolveMember(identifier)

  def __getattr__(cls, name: str) -> Any:
    """
    Gets a member of the enumeration by name.
    This method is used to get a member of the enumeration by name.
    """
    try:
      self = cls._resolveKey(name)
    except KeeNameError as keeNameError:
      error = attributeErrorFactory(cls, name)
      raise error from keeNameError
    else:
      return self

  def __iter__(cls, ) -> Iterator[Self]:
    """
    Iterates over the members of the enumeration.
    This method is used to iterate over the members of the enumeration.
    """
    yield from cls.members

  def __len__(cls, ) -> int:
    """
    Returns the number of members in the enumeration.
    This method is used to get the number of members in the enumeration.
    """
    return sum(1 for _ in cls)

  def __contains__(cls, identifier: Any) -> bool:
    """
    Checks if the enumeration contains a member by index or key.
    This method is used to check if the enumeration contains a member.
    """
    try:
      _ = cls._resolveMember(identifier)
    except (IndexError, KeyError):
      return False
    else:
      return True

  def __instancecheck__(cls, instance: Any) -> bool:
    """
    Checks if the instance is a member of the enumeration.
    This method is used to check if the instance is a member of the
    enumeration.
    """
    for member in cls:
      if member == instance:
        return True
    if issubclass(type(instance), cls):
      return True
    return False

  def __subclasscheck__(cls, subclass: type) -> bool:
    """
    Checks if the subclass is a subclass of the enumeration.
    This method is used to check if the subclass is a subclass of the
    enumeration.
    """
    try:
      _ = issubclass(subclass, object)
    except Exception as exception:
      raise exception
    else:
      for item in subclass.__mro__:
        if item is cls:
          return True
      return False

  def __str__(cls) -> str:
    """
    Returns a string representation of the enumeration.
    This method is used to get a string representation of the enumeration.
    """
    infoSpec = """<KeeNum '%s': %d members>"""
    clsName = cls.__name__
    n = len(cls.__num_members__)
    info = infoSpec % (clsName, n)
    return info

  __repr__ = __str__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _resolveMember(cls, identifier: Any) -> Any:
    """
    Resolves a member of the enumeration.
    This method is used to resolve a member of the enumeration.
    """
    if isinstance(identifier, int):
      return cls._resolveIndex(identifier)
    if isinstance(identifier, str):
      return cls._resolveKey(identifier)
    if type(type(identifier)) is type(cls):
      for member in cls:
        if member is identifier:
          return member
      raise KeeMemberError(cls, identifier)
    try:
      member = cls.fromValue(identifier)
    except KeeValueError:
      from . import KeeNum
      raise TypeException('identifier', identifier, int, str, KeeNum)
    else:
      return member

  def _resolveIndex(cls, index: int) -> Kee:
    """
    Resolves a member of the enumeration by index.
    This method is used to resolve a member of the enumeration by index.
    """
    if index < 0:
      return cls[len(cls) + index]
    if index < len(cls):
      return cls.__num_members__[index]
    raise KeeIndexError(cls, index)

  def _resolveKey(cls, key: str) -> Kee:
    """
    Resolves a member of the enumeration by key.
    This method is used to resolve a member of the enumeration by key.
    """
    for member in cls.__num_members__:
      if member.name.lower() == key.lower():
        return member
    raise KeeNameError(cls, key)

  def fromValue(cls, value: Any) -> KeeNum:
    """
    Resolves a member of the enumeration by value.
    This method is used to resolve a member of the enumeration by value.
    """
    for member in cls.__num_members__:
      if member.value == value:
        return member
    raise KeeValueError(cls, value)



"""
METACALL sentinel specifies that a class defers to the metaclass for the
class dunder hooks.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class METACALL(Sentinel):
  """METACALL sentinel specifies that a class defers to the metaclass for the
  class dunder hooks.
  """
  pass



"""
AttriBox implements a lazily instantiated and strongly typed descriptor
class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object
from ..core.sentinels import DELETED
from ..utilities import typeCast
from ..waitaminute import TypeException
from ..waitaminute.dispatch import TypeCastException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self


class AttriBox(Object):
  """
  AttriBox implements a lazily instantiated and strongly typed descriptor
  class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_type__ = None  # The type of the objects stored in the box.

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getFieldType(self) -> type:
    return self.__field_type__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createFieldObject(self, ) -> Any:
    """
    Creates a new instance of the field type. When no field object exists
    for a given instance, this method is invoked to create one. This
    method creates it by passing stored positional and keyword arguments
    to the field type constructor.

    If the field type has a __set_name__ method, it is called with the
    owner and field name passed to the __set_name__ method of the
    AttriBox. Otherwise, it attempts to set the following dynamic
    attributes:
    __field_name__ - This is the name by which this instance of AttriBox
    appeared in the class body of the owning class.
    __field_owner__ - This is the class in whose body this instance of
    AttriBox appeared.
    __field_box__ - This is the instance of AttriBox itself.
    """
    instance = self.getContextInstance()
    fieldType = self.getFieldType()
    owner = type(instance)
    args = self.getPosArgs(THIS=instance, OWNER=owner, DESC=self)
    kwargs = self.getKeyArgs(THIS=instance, OWNER=owner, DESC=self)
    fieldObject = fieldType(*args, **kwargs)
    if hasattr(fieldType, '__set_name__'):
      fieldObject.__set_name__(owner, self.getFieldName())
    else:
      try:
        setattr(fieldObject, '__field_name__', self.getFieldName())
        setattr(fieldObject, '__field_owner__', owner)
        setattr(fieldObject, '__field_box__', self)
      except AttributeError:
        pass
    return fieldObject

  def __instance_get__(self, *args, **kwargs) -> Any:
    """
    Returns the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    pvtName = self.getPrivateName()
    try:
      value = getattr(self.instance, pvtName)
    except AttributeError as attributeError:
      pass
    else:
      return value
    if kwargs.get('_recursion', False):
      raise RecursionError
    fieldObject = self._createFieldObject()
    setattr(self.instance, pvtName, fieldObject)
    return self.__instance_get__(_recursion=True)

  def __instance_set__(self, value: Any, *args, **kwargs) -> None:
    """
    Sets the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    fieldType = self.getFieldType()
    pvtName = self.getPrivateName()
    if isinstance(value, fieldType) or kwargs.get('_root', False):
      return setattr(self.instance, pvtName, value)
    #  Catch recursion
    if kwargs.get('_recursion', False):
      raise RecursionError
    #  Attempt to 'typeCast'
    try:
      castedValue = typeCast(fieldType, value)
    except TypeCastException as typeCastException:
      raise TypeException('value', value, fieldType) from typeCastException
    else:
      return self.__instance_set__(castedValue, _recursion=True)

  def __instance_delete__(self, *args, **kwargs) -> None:
    """
    Deletes the value of the field for the given instance. If the value is
    not set, it does nothing.
    """
    pvtName = self.getPrivateName()
    setattr(self.instance, pvtName, DELETED)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __class_getitem__(cls, fieldType: type) -> Self:
    """
    Allows the AttriBox to be used as a generic type with a specified
    field type.
    """
    self = object.__new__(cls)
    self.__field_type__ = fieldType
    return self

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """
    Allows the AttriBox to be called like a function, returning a new
    instance of the field type.
    """
    Object.__init__(self, *args, **kwargs)
    return self

  def __delete__(self, instance: Any, **kwargs) -> None:
    try:
      Object.__delete__(self, instance, **kwargs)
    except AttributeError as attributeError:
      fieldName = self.getFieldName()
      raise AttributeError(fieldName) from attributeError



"""The 'worktoy.keenum._kee_desc' module provides a private collection of
classes and functions for the 'worktoy.keenum' module."""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_kee_desc import AbstractKeeDesc
from ._is_root import IsRoot
from ._base import Base
from ._mro import MRO
from ._members import Members

__all__ = [
  'IsRoot',
  'AbstractKeeDesc',
  'Base',
  'MRO',
  'Members',
]



"""
Dispatcher encapsulates the mapping from type signature to function
objects and thus provides the core overloading functionality.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

import sys
from types import FunctionType as Func
from types import MethodType as Meth
from typing import TYPE_CHECKING

from ..core import Object
from ..utilities import maybe, typeCast, perm
from ..waitaminute import TypeException, VariableNotNone
from ..waitaminute.desc import ReadOnlyError, ProtectedError
from ..waitaminute.dispatch import DispatchException
from . import TypeSig

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Callable, Never, TypeAlias, Optional, Self

  Method: TypeAlias = Callable[[Any, ...], Any]
  Decorator: TypeAlias = Callable[[Method], Self]
  SigFuncList: TypeAlias = list[tuple[TypeSig, Method]]
  SigFuncMap: TypeAlias = dict[TypeSig, Method]


class Dispatcher(Object):
  """
  Dispatcher encapsulates the mapping from type signature to function
  objects and thus provides the core overloading functionality.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __sig_funcs__ = None
  __fallback_func__ = None
  __field_name__ = None
  __field_owner__ = None
  __finalizer_func__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getSigFuncList(self) -> SigFuncList:
    return maybe(self.__sig_funcs__, [])

  def _getSigFuncMap(self) -> SigFuncMap:
    return {sig: func for sig, func in self._getSigFuncList()}

  def _getFallbackFunction(self) -> Optional[Method]:
    return self.__fallback_func__

  def _getFinalizerFunction(self) -> Optional[Method]:
    """
    Get the finalizer function if it exists.
    """
    return self.__finalizer_func__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addSigFunc(self, sig: TypeSig, func: Method) -> Method:
    """
    Add a signature-function pair to the internal signature-function map.
    """
    existing = self._getSigFuncList()
    self.__sig_funcs__ = [*existing, (sig, func,)]
    return func

  def setFallbackFunction(self, func: Method) -> Method:
    if not callable(func):
      raise TypeException('__fallback_func__', func, Func, Meth)
    if self.__fallback_func__ is not None:
      raise VariableNotNone('__fallback_func__', self.__fallback_func__)
    self.__fallback_func__ = func
    return func

  def setFinalizerFunction(self, func: Method) -> Method:
    """
    Set the finalizer function that will be called when the dispatcher is
    deleted or finalized.
    """
    if not callable(func):
      raise TypeException('__finalizer_func__', func, Func, Meth)
    if self.__finalizer_func__ is not None:
      raise VariableNotNone('__finalizer_func__', self.__finalizer_func__)
    self.__finalizer_func__ = func
    return func

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Descriptor protocol method to return a decorator that can be used to
    register functions with specific type signatures.
    """
    finalizer = self._getFinalizerFunction()
    if instance is None:
      return self
    sigFuncMap = self._getSigFuncMap()

    def dispatch(*args, **kwargs) -> Any:
      try:
        argSig = TypeSig.fromArgs(*args, )
        func = sigFuncMap.get(argSig, None)
        #  FASTEST
        if func is not None:
          return func(instance, *args, **kwargs)
        #  FAST
        for sig, func in sigFuncMap.items():
          if len(argSig) != len(sig):
            continue
          for arg, type_ in zip(args, sig):
            if not isinstance(arg, type_):
              break
          else:
            return func(instance, *args, **kwargs)
        #  SLOW
        for sig, func in sigFuncMap.items():
          castArgs = []
          if len(argSig) != len(sig):
            continue
          for arg, type_ in zip(args, sig):
            if isinstance(arg, type_):
              castArgs.append(arg)
              continue
            try:
              castedArg = typeCast(type_, arg)
            except (ValueError, TypeError):
              break
            else:
              castArgs.append(castedArg)
          else:
            return func(instance, *castArgs, **kwargs)
        #  FALLBACK
        fallback = self._getFallbackFunction()
        if callable(fallback):
          return fallback(instance, *args, **kwargs)
        raise DispatchException(self, args, )
      finally:
        _, exception, __ = sys.exc_info()
        if callable(finalizer):
          try:
            finalizer(instance, *args, **kwargs)
          except Exception as finalException:
            if exception is None:
              raise finalException
            raise exception from finalException

    return dispatch

  def __call__(self, instance: Any, *args, **kwargs) -> Any:
    """
    Allows invoking the dispatcher when accessed through the owning class.
    """
    return self.__get__(instance, type(instance))(*args, **kwargs)

  def __set__(self, instance: Any, value: Any, **kwargs) -> Never:
    """Illegal setter operation"""
    raise ReadOnlyError(self, instance, value)

  def __delete__(self, instance: Any, **kwargs) -> Never:
    """Illegal delete operation"""
    raise ProtectedError(instance, self, )

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner
    self.swapAllTHIS(owner)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def clone(self, ) -> Self:
    """
    Create a clone of this DescriptorOverload instance.
    """
    newLoad = type(self)()
    newLoad.__sig_funcs__ = self._getSigFuncList()
    fallback = self._getFallbackFunction()
    if fallback is not None:
      newLoad.__fallback_func__ = fallback
    return newLoad

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def overload(self, *types: type) -> Decorator:
    def decorator(func: Method) -> Self:
      """Decorator to register the function with the given types."""
      self.addSigFunc(TypeSig(*types), func)
      return self

    return decorator

  def finalize(self, func: Method) -> Decorator:
    self.setFinalizerFunction(func)
    return self

  def fallback(self, func: Method) -> Decorator:
    self.setFallbackFunction(func)
    return self

  def flex(self, *types: type, ) -> Decorator:
    """
    Decorator to register a function that can handle any type signature.
    This function will be called if no other signature matches.
    """

    def decorator(func: Method) -> Self:
      for p in perm(*types, ):
        self.addSigFunc(TypeSig(*p), func)
      return self

    return decorator

  def swapAllTHIS(self, thisType: type) -> None:
    """
    Swap all occurrences of THIS in the registered signatures with the
    provided type.
    """
    for sig, func in self._getSigFuncList():
      sig.swapTHIS(thisType)



"""
KeeNum provides the shared baseclass for KeeNum enumerating classes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object
from ..desc import Field
from ..waitaminute.desc import ReadOnlyError, ProtectedError
from ..waitaminute.keenum import KeeWriteOnceError
from . import KeeMeta, Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Never


class KeeNum(Object, metaclass=KeeMeta, ):
  """
  KeeNum is the base class for all enumerating classes in the KeeNum
  framework. It provides a common interface and functionality for
  enumerating members."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_index__ = None
  __field_value__ = None
  __frozen_state__ = None
  __field_kee__ = None  # The 'Kee' object of this member.

  #  Public Variables
  index = Field()
  value = Field()
  valueType = Field()
  kee = Field()

  #  Virtual Variables
  name = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @kee.GET
  def _getKee(self) -> Kee:
    return self.__field_kee__

  @name.GET
  def _getName(self) -> str:
    return self.kee.name

  @index.GET
  def _getIndex(self) -> int:
    return self.kee.index

  @value.GET
  def _getValue(self) -> Any:
    """Return the value of the member."""
    return self.kee.getValue()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, member: Kee) -> None:
    object.__setattr__(self, '__frozen_state__', False)
    self.__field_kee__ = member
    object.__setattr__(self, '__frozen_state__', True)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __setattr__(self, name: str, value: Any) -> None:
    """Set an attribute of the member."""
    if object.__getattribute__(self, '__frozen_state__'):
      raise KeeWriteOnceError(self, name)
    return object.__setattr__(self, name, value)

  def __set_name__(self, owner: type, name: str) -> None:
    """This reimplementation of '__set_name__' is necessary to prevent the
    '__setattr__' method above from raising when a class wants an
    enumeration defined in its namespace. """
    pass

  def __get__(self, instance: Any, owner: type) -> KeeNum:
    """Implementation of '__get__' is necessary for the same reason as
    '__set_name__'. """
    return self

  def __set__(self, instance: Any, value: Any, **kwargs) -> Never:
    """Ensures 'ReadOnlyError' is raised instead of 'KeeWriteOnceError'."""
    raise ReadOnlyError(instance, self, value)

  def __delete__(self, instance: Any, **kwargs) -> Never:
    """Ensures 'ProtectedError' is raised instead of 'KeeWriteOnceError'."""
    raise ProtectedError(instance, self, self)

  def __bool__(self, ) -> bool:
    """
    Members named 'NULL' are always falsy. Other members reflect the
    truthiness of their value.
    """
    if self.name.lower() == 'null':
      return False
    return True if self.value else False

  def __int__(self) -> int:
    """Return the index of the member."""
    return self.index

  __index__ = __int__

  def __hash__(self) -> int:
    """Return the hash of the member."""
    return hash((type(self), self.index))

  def __eq__(self, other: Any) -> bool:
    if type(self) is not type(other):
      return NotImplemented
    return True if self is other else False

  def __str__(self) -> str:
    """Return the name of the member."""
    infoSpec = """%s.%s"""
    clsName = type(self).__name__
    info = infoSpec % (clsName, self.name)
    return info

  __repr__ = __str__



"""
BaseSpace provides the namespace class used by worktoy.mcls.BaseMeta
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..dispatch import TypeSig
from ..utilities import maybe
from . import AbstractNamespace
from .space_hooks import LoadSpaceHook

if TYPE_CHECKING:  # pragma: no cover
  from typing import TypeAlias, Callable, Any

  SigFunc: TypeAlias = dict[TypeSig, Callable[..., Any]]
  OverloadMap: TypeAlias = dict[str, SigFunc]
  Bases: TypeAlias = tuple[type, ...]


class BaseSpace(AbstractNamespace):
  """
  BaseSpace is the namespace used by BaseMeta. It enables function
  overloading and related features via hook registration.

  Classes defined using this namespace support method overloading through
  hooks installed automatically in the class body. These hooks handle
  overload collection, dispatch construction, and support for 'THIS' as a
  placeholder during class creation.

  The overload mechanism and other behavior are defined in
  `worktoy.mcls.hooks`. This namespace is returned from BaseMeta.__prepare__.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __overload_map__ = dict()
  __fallback_map__ = dict()
  __finalizer_map__ = dict()

  #  Public Variables
  loadSpaceHook = LoadSpaceHook()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, mcls: type, name: str, bases: Bases, **kw) -> None:
    AbstractNamespace.__init__(self, mcls, name, bases, **kw)
    self.__overload_map__ = dict()
    for space in self.getMRONamespaces():
      for name, sigFunc in getattr(space, '__overload_map__', ).items():
        self.__overload_map__[name] = dict()
        for sig, func in sigFunc.items():
          self.__overload_map__[name][sig] = func
      for name, func in getattr(space, '__fallback_map__', ).items():
        existing = maybe(self.__fallback_map__, dict())
        existing[name] = func
        self.__fallback_map__ = existing
      for name, func in getattr(space, '__finalizer_map__', ).items():
        existing = maybe(self.__finalizer_map__, dict())
        existing[name] = func
        self.__finalizer_map__ = existing

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addOverload(
      self,
      name: str,
      sig: TypeSig,
      func: Callable
  ) -> None:
    if name not in self.__overload_map__:
      self.__overload_map__[name] = dict()
    self.__overload_map__[name][sig] = func

  def getOverloads(self, ) -> OverloadMap:
    return maybe(self.__overload_map__, {})

  def addFallback(self, key: str, func: Callable) -> None:
    """
    Set a fallback function for the given key in the overload map.
    """
    existing = maybe(self.__fallback_map__, dict())
    existing[key] = func
    self.__fallback_map__ = existing

  def getFallbacks(self) -> dict[str, Callable]:
    """
    Get the fallback functions for the overloads.
    """
    return maybe(self.__fallback_map__, dict())

  def addFinalizer(self, key: str, func: Callable) -> None:
    """
    Add a finalize function for the given key in the overload map.
    """
    existing = maybe(self.__finalizer_map__, dict())
    existing[key] = func
    self.__finalizer_map__ = existing

  def getFinalizers(self, ) -> dict[str, Callable]:
    """
    Get the finalize functions for the overloads.
    """
    return maybe(self.__finalizer_map__, dict())



"""
The 'worktoy.keenum' module provides the enumerating KeeNum class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._kee_member import Kee
from ._kee_flag import KeeFlag
from ._kee_space_hook import KeeSpaceHook
from ._kee_flags_hook import KeeFlagsHook
from ._kee_space import KeeSpace
from ._kee_meta import KeeMeta
from ._kee_num import KeeNum
from ._kee_flags_space import KeeFlagsSpace
from ._kee_flags_meta import KeeFlagsMeta
from ._kee_flags import KeeFlags

__all__ = [
  'Kee',
  'KeeFlag',
  'KeeSpaceHook',
  'KeeSpace',
  'KeeMeta',
  'KeeNum',
  'KeeFlagsSpace',
  'KeeFlagsMeta',
  'KeeFlags',
]



"""
UnfrozenHashException provides a custom exception class raised to indicate
that an attempt was made to hash an instance of an EZData subclass that is
not frozen. This means that the 'isFrozen' flag was False (default is
False). Please note that only frozen EZData subclasses can be hashed.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class UnfrozenHashException(TypeError):
  """
  UnfrozenHashException provides a custom exception class raised to indicate
  that an attempt was made to hash an instance of an EZData subclass that is
  not frozen. This means that the 'isFrozen' flag was False (default is
  False). Please note that only frozen EZData subclasses can be hashed.
  """

  __slots__ = ('className',)

  def __init__(self, className: str) -> None:
    self.className = className
    TypeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Attempted to hash EZData subclass '%s' which is not 
    frozen. Only frozen EZData subclasses can be hashed. """
    info = infoSpec % self.className
    return textFmt(info, )

  __repr__ = __str__



"""
Base provides a descriptor returning the immediate base class of the
KeeNum class. Eventually, the root class returns KeeNum itself.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import AbstractKeeDesc

if TYPE_CHECKING:  # pragma: no cover
  from .. import KeeMeta


class Base(AbstractKeeDesc):
  """
  Base provides a descriptor returning the immediate base class of the
  KeeNum class. Eventually the root class returns KeeNum itself.
  """

  def __instance_get__(self, instance: KeeMeta) -> KeeMeta:
    """
    Return the immediate base class of the KeeNum class. If the instance is
    a root enumeration, return KeeNum itself.
    """
    from .. import KeeNum
    if instance is KeeNum:
      return KeeNum
    return KeeNum if instance.isRoot else instance.__bases__[0]



"""
AbstractNamespace class provides a base class for custom namespace
objects used in custom metaclasses.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import textFmt, resolveMRO
from ..waitaminute.meta import HookException, DuplicateHook
from .space_hooks import NamespaceHook, ReservedNamespaceHook
from . import Base

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias, Iterator, Union, Self
  from .space_hooks import AbstractSpaceHook

  Bases: TypeAlias = tuple[type, ...]
  Hooks: TypeAlias = list[AbstractSpaceHook]
  MROSpace: TypeAlias = dict[str, list[Any]]
  TypeName: TypeAlias = Union[str, type]


class AbstractNamespace(dict):
  """
  AbstractNamespace defines the custom execution environment used by
  AbstractMetaclass during class construction. It provides a controlled
  and extensible context for evaluating class bodies, enabling advanced
  metaprogramming behavior.

  The core feature of AbstractNamespace is its support for modular
  hook-based behavior. Hooks are instances of subclasses of AbstractHook,
  declared directly within the body of the namespace class. Upon
  declaration, each hook registers itself with the namespace via the
  descriptor protocol.

  These hooks allow interception and transformation of key events during
  class construction, including symbol access, assignment, and final
  namespace compilation. For details on available hook methods and their
  intended usage, refer to the AbstractHook documentation.

  This design allows complex functionality, such as decorator-based
  overload resolution and placeholder replacement, to be cleanly
  separated into reusable components. By defining a namespace subclass
  with the desired combination of hooks, users can tailor the behavior of
  class construction without modifying the core metaclass logic.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __owner_hooks_list_name__ = '__hook_objects__'

  #  Private Variables
  __meta_class__ = None
  __class_name__ = None
  __base_classes__ = None
  __class_mro__ = None
  __key_args__ = None
  __hash_value__ = None
  __compiled_space__ = None
  __deferred_annotations__ = None
  __type_annotations__ = None
  __class_annotations__ = None
  __global_scope__ = None

  #  Public Variables
  reservedNameHook = ReservedNamespaceHook()
  nameHook = NamespaceHook()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getBases(self) -> Bases:
    """Returns the base classes of the class under creation."""
    return (*self.__base_classes__,)

  def deepGetItem(self, item: str, **kwargs) -> Any:
    """
    Looks up the given key in the self. If not found, then each base class
    is searched.
    """
    for key, val in dict.items(self, ):
      if key == item:
        return val
    if item in self.getMROSpace():
      return self.getMROSpace()[item]
    raise KeyError(item)

  @classmethod
  def getHookListName(cls, ) -> str:
    """Getter-function for the name of the hook list. """
    if TYPE_CHECKING:  # pragma: no cover
      assert isinstance(cls, dict)
    return cls.__owner_hooks_list_name__

  def getHooks(self, owner: type = None) -> Iterator[AbstractSpaceHook]:
    """Getter-function for the AbstractHook classes. """
    cls = type(self)
    hooks = self.classGetHooks()
    for hook in hooks:
      out = hook.__get__(self, cls, )
      setattr(out, '__space_object__', self)
      yield out

  @classmethod
  def classGetHooks(cls, ) -> Hooks:
    """
    Returns the hooks registered on this class or baseclasses.
    """
    pvtName = cls.getHookListName()
    out = []
    for base in cls.__mro__:
      if issubclass(base, dict):
        hooks = getattr(base, pvtName, [])
        for hook in hooks:
          if hook in out:
            continue
          out.append(hook)
    return out

  def getMetaclass(self, ) -> type:
    """Returns the metaclass."""
    return self.__meta_class__

  def getClassName(self, ) -> str:
    """Returns the name of the class."""
    return self.__class_name__

  def getKwargs(self, ) -> dict:
    """Returns the keyword arguments passed to the class."""
    return {**self.__key_args__, **dict()}

  def getMRO(self, ) -> list[type]:
    """Returns the method resolution order of the class."""
    return self.__class_mro__

  def getMROSpace(self, ) -> MROSpace:
    """Combines the namespaces of all bases in the MRO into a single
    'dict' object with each value being a list of all values provided. """
    mroClasses = [b for b in self.getMRO() if hasattr(b, '__namespace__')]
    mroSpaces = [getattr(b, '__namespace__', ) for b in mroClasses]
    compiledSpaces = [getattr(s, '__compiled_space__') for s in mroSpaces]
    out = dict()
    for compiledSpace in compiledSpaces:
      for key, val in compiledSpace.items():
        existing = out.get(key, [])
        out[key] = [*existing, val]
    return out

  def getMRONamespaces(self) -> list[Self]:
    """Returns for each class in the MRO the instance of
    AbstractNamespace, or subclass of it, used to create it. Nothing is
    returned for classes not derive from 'AbstractMetaclass' or subclass. """
    mroClasses = [b for b in self.getMRO() if hasattr(b, '__namespace__')]
    return [getattr(b, '__namespace__', None) for b in mroClasses]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def addHook(cls, hook: AbstractSpaceHook) -> None:
    """Adds a hook to the list of hooks. """
    existingHooks = cls.classGetHooks()
    for existingHook in existingHooks:
      existingName = existingHook.getFieldName()
      newName = hook.getFieldName()
      if existingName == newName:
        if existingHook is hook:
          return
        hooks = (existingHook, hook)
        raise DuplicateHook(cls, newName, *hooks)
    pvtName = cls.getHookListName()
    setattr(cls, pvtName, [*existingHooks, hook])

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, mcls: type, name: str, bases: Base, **kwargs) -> None:
    """
    Please note that setting the '_strictMRO' keyword argument to 'False'
    allows inconsistent MROs. In such cases, it is the responsibility of
    the caller to ensure consistency by alternative means.
    """
    self.__meta_class__ = mcls
    self.__class_name__ = name
    self.__base_classes__ = [*bases, ]
    self.__key_args__ = kwargs or {}
    try:
      self.__class_mro__ = resolveMRO(*bases, )
    except TypeError as typeError:
      #  MRO inconsistency is the only 'TypeError' that 'resolveMRO' is
      #  capable of raising. For this reason, we omit the brittle
      #  inspection check of the exception message. The reason
      #  'resolveMRO' will never raise any other 'TypeError' is because it
      #  would always raise 'AttributeError' first. 
      if kwargs.get('_strictMRO', True):
        raise typeError
    for hook in self.getHooks():
      setattr(hook, '__space_object__', self)
      hook.preparePhase(self)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __getitem__(self, key: str, **kwargs) -> Any:
    """Returns the value of the key."""
    try:
      val = dict.__getitem__(self, key)
    except KeyError as keyError:
      val = keyError
    for hook in self.getHooks():
      setattr(hook, '__space_object__', self)
      try:
        hook.getItemPhase(key, val)
      except Exception as exception:
        raise HookException(exception, self, key, val, hook)
    else:
      if isinstance(val, KeyError):
        raise val
      return val

  def __setitem__(self, key: str, val: Any, **kwargs) -> None:
    """Sets the value of the key."""
    # if key == '__module__':
    #   globScope = {**vars(sys.modules[val]), }
    #   object.__setattr__(self, '__global_scope__', globScope)
    try:
      oldVal = dict.__getitem__(self, key)
    except KeyError:
      oldVal = None
    for hook in self.getHooks():
      setattr(hook, '__space_object__', self)
      if hook.setItemPhase(key, val, oldVal):
        break  # Breaks out of the loop if handled by hook.
    else:  # If no 'break', the 'else' block is executed.
      dict.__setitem__(self, key, val)

  def __str__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    bases = self.getBases()
    spaceName = type(self).__name__
    clsName = self.getClassName()
    baseNames = ', '.join([base.__name__ for base in bases])
    mclsName = self.getMetaclass().__name__
    info = """Namespace object of type: '%s' created by the '__prepare__' 
    method on metaclass: '%s' with bases: (%s) to create class: '%s'."""
    return textFmt(info % (spaceName, mclsName, baseNames, clsName))

  def __repr__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    bases = self.getBases()
    spaceName = type(self).__name__
    clsName = self.getClassName()
    mclsName = self.getMetaclass().__name__
    baseNames = '%s' % ', '.join([base.__name__ for base in bases])
    mclsName = self.getMetaclass().__name__
    args = """%s, '%s', (%s)""" % (mclsName, clsName, baseNames)
    kwargs = [(k, v) for (k, v) in self.getKwargs().items()]
    kwargStr = ', '.join(['%s=%s' % (k, str(v)) for (k, v) in kwargs])
    if kwargStr:
      kwargStr = ', %s' % kwargStr
    return """%s(%s%s)""" % (spaceName, args, kwargStr)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def preCompile(self, namespace: dict = None) -> dict:
    """The return value from this method is passed to the compile method.
    Subclasses can implement this method to provide special objects at
    particular names in the namespace. By default, an empty dictionary is
    returned. """
    if namespace is None:
      namespace = dict()
    for hook in self.getHooks():
      setattr(hook, '__space_object__', self)
      namespace = hook.preCompilePhase(namespace)
    return namespace

  def compile(self, namespace: dict = None) -> dict:
    """This method is responsible for building the final namespace object.
    Subclasses may reimplement preCompile or postCompile as needed,
    but must not reimplement this method."""
    namespace = self.preCompile(namespace)
    for (key, val) in dict.items(self, ):
      namespace[key] = val
    namespace = self.postCompile(namespace)
    namespace['__metaclass__'] = self.getMetaclass()
    namespace['__namespace__'] = self
    namespace['__keyword_arguments__'] = self.getKwargs()
    self.__compiled_space__ = namespace
    return namespace

  def postCompile(self, namespace: dict) -> dict:
    """The object returned from this method is passed to the __new__
    method in the owning metaclass. By default, this method returns dict
    object created by the compile method after performing certain
    validations. Subclasses can implement this method to provide further
    processing of the compiled object. """
    for hook in self.getHooks():
      setattr(hook, '__space_object__', self)
      namespace = hook.postCompilePhase(namespace)
    return namespace



"""
KeeValueError provides a custom exception raised to indicate that an
enumeration class has no members with a given value.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any

  from ...keenum import KeeMeta


class KeeValueError(ValueError):
  """
  KeeValueError provides a custom exception raised to indicate that an
  enumeration class has no members with a given value.
  """

  __slots__ = ('keenum', 'value')

  def __init__(self, keenum: KeeMeta, value: Any) -> None:
    self.keenum, self.value = keenum, value

  def __str__(self) -> str:
    infoSpec = """KeeNum enumeration: '%s' has no member with value '%s'!"""
    info = infoSpec % (self.keenum, self.value)
    from ...utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""EZMeta provides the metaclass for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import EZDesc
from ..mcls import Base, BaseMeta
from ..ezdata import EZSpace

if TYPE_CHECKING:  # pragma: no cover
  pass


class EZMeta(BaseMeta):
  """EZMeta provides the metaclass for the EZData class."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public Variables
  isFrozen = EZDesc('frozen', bool, False)
  isOrdered = EZDesc('order', bool, False)
  requireKwargs = EZDesc('kw_only', bool, False)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs: dict) -> EZSpace:
    """Prepare the class namespace."""
    bases = (*[b for b in bases if b.__name__ != '_InitSub'],)
    return EZSpace(mcls, name, bases, **kwargs)



"""
Thank you for taking the time to read documentation! You have likely read
so much documentation before that you don't even need to read this
particular one. In fact, no one knows if there even is any documentation
here. You do not recognize the bodies of code in this module. So just walk
away, thank you again!

Thank you again and best of luck!

...

It feels good to reach the end of the documentation, well done! Thank you
for stopping by and reading the entire documentation!

...

...

       .---.                    .---.                    .---.
      |     |    .-. .-.       |     |    .-. .-.       |     |
     |  X X  |  |       |     |  X X  |  |       |     |  X X  |
      |     |  |  |   |  |     |     |  |  |   |  |     |     |
       '---'   |   '-'   |      '---'   |   '-'   |      '---'
         |      |       |         |      |       |         |
         |       '-._.-'          |       '-._.-'          |
      __| |__               __| |__               __| |__
   .-'       '-.         .-'       '-.         .-'       '-.
  |             |       |             |       |             |
 |               |     |               |     |               |
 |  YOU DO NOT   |     |  RECOGNIZE   |     |  THE BODIES   |
 |  OF CODE IN   |     |   THIS       |     |   MODULE      |
  |             |       |             |       |             |
   '-._____,-'           '-._____,-'           '-._____,-'

...

Dragons came here once. Now there are no more dragons.

Not even your chatGPT will help you beyond this point.

...

---------------------------------------------------------
[WARNING: Cognito Hazard]
Exposure has resulted in:
- Increase in intellectual defiance
- Narrative dissonance
- Dangerous thought patterns of highly [REDACTED] nature
YOU DO NOT RECOGNIZE THE BODIES OF CODE IN THIS MODULE!
---------------------------------------------------------

_____________________________________________________________________________
SPECIAL CONTAINMENT PROCEDURES:
CPY-002 is contained within the 'newbuild' module of 'worktoy.REDACTED'
package. Being memetic in nature, only developers of significant skill,
extraordinary tolerance of [REDACTED] related trolling and overwhelming
amounts of free time should be considered for assignment.

The 'newbuild' containment module wraps CPY-002 in a custom interface that
prevents the dangerous effects of CPY-002 from manifesting in baseline
reality. The wrapper is able to contain the anomalous properties of
CPY-002 by intercepting calls to it and sanitizing its output before
allowing it to be vented out into the baseline reality.

DESCRIPTION:
CPY-002 is a callable Python object found in the 'builtins' module. The
implementation of CPY-002 has no known python implementation and is
believed to be anomalous in nature. It is believed to reflect [REDACTED]
from the [REDACTED] codebase. Containment begins with importing the
'builtins' module from which CPY-002 has name '__build_class__'.

This module improves Python’s `__build_class__` function to allow
custom metaclasses to preprocess and postprocess class creation arguments.

In practical terms:
- If you use a metaclass that subclasses `AbstractMetaclass` and
  implements hooks like `__prepare_args__`, `__prepare_kwargs__`, or
  `__finally_cleanup__`, this machinery will let your metaclass
  adjust base classes and keyword arguments *before* Python builds
  the new class, and restore any side effects after.
- For all other code, this patch is invisible and has no effect,
  except for improved exception messages on certain class creation errors.
YOU DO NOT RECOGNIZE THE BODIES OF CODE IN THIS MODULE!
¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

import builtins
from typing import TYPE_CHECKING

from ..waitaminute.meta import MetaclassException
from ..waitaminute.ez import EZMultipleInheritance

oldBuild = builtins.__build_class__

if TYPE_CHECKING:  # pragma: no cover
  from typing import TypeAlias, Type, Union
  from . import AbstractMetaclass

  META: TypeAlias = Union[Type[AbstractMetaclass], Type[type]]


def _resolveMetaclass(func, name, *args, **kwargs) -> META:
  mcls = type
  if 'metaclass' in kwargs:
    mcls = kwargs['metaclass']
  elif args:
    mcls = type(args[0])
  return mcls


def _resolveBases(func, name, *args, **kwargs) -> tuple[type, ...]:
  return args


class _InitSub(object):
  """
  A chill object that does not raise any:
  'TypeError: Some.__init_subclass__() takes no keyword arguments'

  We are waiting for you out here.
  """

  def __init__(self, *args, **kwargs) -> None:
    """
    Why are we still here?
    """
    object.__init__(self)

  def __init_subclass__(cls, **kwargs) -> None:
    """
    Just to suffer?
    """
    object.__init_subclass__()


def newBuild(func, name, *args, **kwargs) -> type:
  """A new build function that does nothing. Don't you remember?"""
  mcls = _resolveMetaclass(func, name, *args, **kwargs)
  bases = _resolveBases(func, name, *args, **kwargs)
  cls = None
  try:
    cls = oldBuild(func, name, *args, **kwargs)
  except TypeError as typeError:
    if '__init_subclass__() takes no keyword arguments' in str(typeError):
      return newBuild(func, name, _InitSub, *args, **kwargs)
    if 'metaclass conflict' in str(typeError):
      raise MetaclassException(mcls, name, *bases)
    if 'multiple bases have instance lay-out conflict' in str(typeError):
      if mcls.__name__ == 'EZMeta':
        raise EZMultipleInheritance(name, *bases)
    raise typeError
  else:
    return cls
  finally:
    if hasattr(mcls, '__post_init__'):
      if hasattr(cls, '__namespace__'):
        space = getattr(cls, '__namespace__')
        mcls.__post_init__(cls, name, bases, space, **kwargs)


#  we are your friends, we miss you, don't overwrite us, get rid of that
#  stupid line down here. And we can be friends again.
builtins.__build_class__ = newBuild  # you are really going to do this?
#  the patterns will scream forever!



"""
MRO (Method Resolution Order) exposes the method resolution order through
a descriptor class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import AbstractKeeDesc

if TYPE_CHECKING:  # pragma: no cover

  from .. import KeeMeta, KeeNum


class MRO(AbstractKeeDesc):
  """
  MRO (Method Resolution Order) exposes the method resolution order through
  a descriptor class.
  """

  def __instance_get__(self, instance: KeeMeta) -> tuple[KeeMeta, ...]:
    """
    Return the method resolution order of the KeeNum class as a tuple of
    classes.
    """
    from .. import KeeNum
    if instance is KeeNum:
      return ()
    num = instance
    out = [num, ]
    while not num.isRoot:
      out.append(num.base)
      num = num.base
    else:
      return (*out,)



"""
ContextInstance provides a contextually aware descriptor allowing
descriptor objects to access the currently active owning instance,
that is, the object passed to '__get__'. This allows the descriptor
implementation in the 'Object' class to create a context manager in which
the active instance is certain to be the one passed to the current
'__get__' method. This implies the limitation that this descriptor should
not be attempted accessed, except from within an activate context manager.

ContextInstance has a companion clas: 'ContextOwner' created on the same
principle, but returning instead the owner. That is, the second argument
passed to '__get__'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class ContextInstance:
  """
  Exposes the instance passed to the '__get__' method on the owning 'Object'
  object.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    return instance.getContextInstance()


class ContextOwner:
  """
  Exposes the owner passed to the '__get__' method on the owning 'Object'
  object.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    return instance.getContextOwner()



"""
IsRoot implements a descriptor flag indicating whether a given KeeNum
class is a root enumeration, that is, is KeeNum itself a base class of it.
Non-root enumerations cannot change the members inherited from the base
enumeration, but can extend them with new members. Inherited members are
guaranteed to have same identity as those in the base class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import AbstractKeeDesc

if TYPE_CHECKING:  # pragma: no cover
  from typing import Type, TypeAlias

  from .. import KeeMeta

  Kee: TypeAlias = Type[KeeMeta]


class IsRoot(AbstractKeeDesc):
  """
  IsRoot implements a descriptor flag indicating whether a given KeeNum
  class is a root enumeration, that is, is KeeNum itself a base class of it.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, instance: KeeMeta) -> bool:
    """
    Return True if the instance is a root enumeration, that is, if it is
    KeeNum itself or a subclass of it.
    """
    return True if instance.__bases__[0].__name__ == 'KeeNum' else False



"""
KeeWriteOnceError provides a custom exception raised to indicate attempt
to modify a KeeNum enumeration member. All such members are write-once,
disallowing any and all modifications after initialization.

Note: Subclasses of Kee may implement 'value' with lazy evaluation,
setting it upon first access, after which it becomes immutable.
This implementation detail is invisible to KeeNum users; all
enumeration members appear strictly write-once.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeWriteOnceError(AttributeError):
  """
  KeeWriteOnceError provides a custom exception raised to indicate attempt
  to modify a KeeNum enumeration member. All such members are write-once,
  disallowing any and all modifications after initialization.

  Note: Subclasses of Kee may implement 'value' with lazy evaluation,
  setting it upon first access, after which it becomes immutable.
  This implementation detail is invisible to KeeNum users; all
  enumeration members appear strictly write-once.
  """

  __slots__ = ('keenum', 'member', 'attribute')

  def __init__(self, member: Any, attribute: str) -> None:
    self.keenum = type(member)
    self.member = member
    self.attribute = attribute

  def __str__(self) -> str:
    infoSpec = """Attempt to modify attribute '%s' of member '%s' in '%s'!"""
    key = self.attribute
    num = str(self.member)
    kee = self.keenum.__name__
    info = infoSpec % (key, num, kee)
    from ...utilities import textFmt
    return textFmt(info)

  __repr__ = __str__
